-- Notes
--
-- UI Frames once created are never released by WoW until a /reload.
-- Therefore, frames are spawned and returned to a pool.

SeedRaid = {}

-- Configuration, UI constants are from UI.xml
SeedRaid.frame = SR
SeedRaid.close = SR_Close

SeedRaid.raidClip  = SR_ScrollFrame
SeedRaid.content   = SR_ScrollFrame_Content
SeedRaid.scrollbar = SR_ScrollFrame_ScrollBar

SeedRaid.seedClip         = SR_ScrollFrame_Planted
SeedRaid.contentPlanted   = SR_ScrollFrame_Planted_Content
SeedRaid.scrollbarPlanted = SR_ScrollFrame_Planted_ScrollBar

SeedRaid.lootClip      = SR_ScrollFrame_Loot
SeedRaid.contentLoot   = SR_ScrollFrame_Loot_Content
SeedRaid.scrollbarLoot = SR_ScrollFrame_Loot_ScrollBar

SeedRaid.debug = false

--
-- Clears current data.
-- @notes
-- WoW scans the global table (_G) for "SLASH_" and uses string matching
-- to connect the slash command with its matching function.
-- 
SLASH_CLEAR1 = "/sr_clear"
function SlashCmdList.CLEAR(msg, editbox)
  if not SeedRaid then
    return
  end
  SeedRaid:Reset()
  print("SeedRaid results cleared.")
end

--
-- Shows the SeedRaid window.
-- @notes
-- WoW scans the global table (_G) for "SLASH_" and uses string matching
-- to connect the slash command with its matching function.
-- 
SLASH_SHOW1 = "/sr_show"
function SlashCmdList.SHOW(msg, editbox)
  if not SeedRaid or SeedRaid.frame then
    return
  end
  SeedRaid.frame:Show()
end

--
-- Adds a raid member (name-realm) to the member table.
-- @param name Raid member name to enter into member table.
-- @param unitId UnitId of the member, used by many WoW API functions.
-- @param rank Raid rank of the member.
-- @return Raid member table.
--
function SeedRaid:AddMember(name, unitId, rank)
  if (not SeedRaid)                               or
     (not SeedRaid.seeds or not SeedRaid.members) or
     (not name     or type(name)   ~= "string")   or
     (not unitId   or type(unitId) ~= "string")   or
     (not rank     or type(rank)   ~= "number")   then
    return nil
  end

  local member = SeedRaid.members[name]
  if member == nil then
    -- Spawn new member
    local frame = SeedRaid:GetCheckButton("raid")
    local seedsCount = {}
    for i, v in pairs(SeedRaid.seeds) do
      seedsCount[i] = 0
    end

    member = {
      frameRoster   = frame,
      inPartyOrRaid = true,
      loots         = {},
      marked        = 0,
      plantedAll    = false,
      rank          = rank,
      seeds         = seedsCount,
      seedsSum      = 0,
      unitId        = unitId
    }

    SeedRaid.members[name] = member
    SeedRaid.memberCount = SeedRaid.memberCount + 1
  else
    member.rank = rank
    -- Update unitId if switched from raid to party, or vice versa
    member.unitId = unitId
  end
  return member
end

--
-- Adds you, the player, to the raid roster
-- @notes
-- The player is always treated differently than a generic raid member,
-- so a lot of extra code exists to handle that.
--
function SeedRaid:AddPlayer()
  if SeedRaid.playerName or SeedRaid.player then
    return
  end

  SeedRaid.playerName = GetUnitName("player", false).."-"..GetRealmName("player")
  SeedRaid.player = SeedRaid:AddMember(SeedRaid.playerName, "player", SeedRaid.RAID_MEMBER)

  -- Default player displayer properties
  SeedRaid.player.frameRoster:SetPoint("TOPLEFT")
end

--
-- Disable events, at times when race conditions are an issue.
--
function SeedRaid:DisableEvents()
  local _ = function(...) end

  -- Disable game events first, as that is where most race conditions would be
  SeedRaid.frame:SetScript("OnEvent", _)

  SeedRaid.frame:SetScript("OnDragStart", _)
  SeedRaid.frame:SetScript("OnDragStop", _)

  SeedRaid.raidClip:SetScript("OnMouseWheel", _)
  SeedRaid.scrollbar:SetScript("OnMouseWheel", _)
  SeedRaid.seedClip:SetScript("OnMouseWheel", _)
  SeedRaid.scrollbarPlanted:SetScript("OnMouseWheel", _)
  SeedRaid.lootClip:SetScript("OnMouseWheel", _)
  SeedRaid.scrollbarLoot:SetScript("OnMouseWheel", _)

  SR_SeedCount:SetScript("OnEnterPressed", _)
  SR_SeedCount:SetScript("OnEscapePressed", _)
  SR_SeedCount:SetScript("OnTabPressed", _)

  SR_RoundSize:SetScript("OnEnterPressed", _)
  SR_RoundSize:SetScript("OnEscapePressed", _)
  SR_RoundSize:SetScript("OnTabPressed", _)

  SR_AlertInterval:SetScript("OnEnterPressed", _)
  SR_AlertInterval:SetScript("OnEscapePressed", _)
  SR_AlertInterval:SetScript("OnTabPressed", _)

  SeedRaid.scrollbar:SetScript("OnValueChanged", _)
  SeedRaid.scrollbarLoot:SetScript("OnValueChanged", _)
end

--
-- Enables frame and game events.
--
function SeedRaid:EnableEvents()
  SeedRaid.frame:SetScript("OnDragStart",
    function(self, button)
      SeedRaid.frame:StartMoving()
    end
  )

  SeedRaid.frame:SetScript("OnDragStop",
    function(self, button)
      SeedRaid.frame:StopMovingOrSizing()
    end
  )

  SeedRaid.raidClip:SetScript("OnMouseWheel", SeedRaid.OnMouseWheel)
  SeedRaid.scrollbar:SetScript("OnMouseWheel", SeedRaid.OnMouseWheel)
  SeedRaid.seedClip:SetScript("OnMouseWheel", SeedRaid.OnMouseWheel)
  SeedRaid.scrollbarPlanted:SetScript("OnMouseWheel", SeedRaid.OnMouseWheel)
  SeedRaid.lootClip:SetScript("OnMouseWheel", SeedRaid.OnMouseWheel)
  SeedRaid.scrollbarLoot:SetScript("OnMouseWheel", SeedRaid.OnMouseWheel)

  SeedRaid.scrollbar:SetScript("OnValueChanged", 
    function (self, value)
      self:GetParent():SetVerticalScroll(value)
      SeedRaid:UpdateRaidDisplay()
    end
  )

  SeedRaid.scrollbarLoot:SetScript("OnValueChanged", 
    function (self, value)
      self:GetParent():SetVerticalScroll(value)
      SeedRaid:UpdateRaidDisplay()
    end
  )

  -- Seed Count
  SR_SeedCount:SetScript("OnEditFocusGained",
    function (self)
      SR_SeedCount:HighlightText()
    end
  )
  SR_SeedCount:SetScript("OnEnterPressed", SeedRaid.SetSeedCount)
  SR_SeedCount:SetScript("OnEscapePressed",
    function (self)
      self:SetText(SeedRaid.seedCount)
      SeedRaid.SetSeedCount(self)
    end
  )
  SR_SeedCount:SetScript("OnTabPressed",
    function (self)
      SeedRaid.SetSeedCount(self)
      if IsShiftKeyDown() then
        SR_AlertInterval:SetFocus()
      else
        SR_RoundSize:SetFocus()
      end
    end
  )

  -- Round Size
  SR_RoundSize:SetScript("OnEditFocusGained",
    function (self)
      SR_RoundSize:HighlightText()
    end
  )
  SR_RoundSize:SetScript("OnEnterPressed", SeedRaid.SetRoundSize)
  SR_RoundSize:SetScript("OnEscapePressed",
    function (self)
      self:SetText(SeedRaid.roundSize)
      SeedRaid.SetRoundSize(self)
    end
  )
  SR_RoundSize:SetScript("OnTabPressed",
    function (self)
      SeedRaid.SetRoundSize(self)
      if IsShiftKeyDown() then
        SR_SeedCount:SetFocus()
      else
        SR_AlertInterval:SetFocus()
      end
    end
  )

  -- Alert Interval
  SR_AlertInterval:SetScript("OnEditFocusGained",
    function (self)
      SR_AlertInterval:HighlightText()
    end
  )
  SR_AlertInterval:SetScript("OnEnterPressed", SeedRaid.SetAlertInterval)
  SR_AlertInterval:SetScript("OnEscapePressed",
    function (self)
      self:SetText(SeedRaid.alertInterval)
      SeedRaid.SetAlertInterval(self)
    end
  )
  SR_AlertInterval:SetScript("OnTabPressed",
    function (self)
      SeedRaid.SetAlertInterval(self)
      if IsShiftKeyDown() then
        SR_RoundSize:SetFocus()
      else
        SR_SeedCount:SetFocus()
      end
    end
  )

  -- Enable game events last, as that is where most race conditions would be
  SeedRaid.frame:SetScript("OnEvent", SeedRaid.OnEvent)
end

--
-- Get a CheckButton widget from the frame pool.
--
function SeedRaid:GetCheckButton(type)
  if type ~= "raid" and type ~= "seed" and type ~= "loot" then
    print(
      "SeedRaid:GetCheckButton",
      "Error, " .. tostring(type) .. " is not a valid framepool."
    )
    return nil
  end

  local pool = SeedRaid.framepool[type]
  local parent = SeedRaid.framepoolParents[type]

  local count = table.getn(pool)
  local frame = table.remove(pool)
  if not frame then
    frame =
    CreateFrame(
      "CheckButton",
      "$parent_CheckButton_" .. count,
      parent,
      "SR_EntryTemplate"
    )
    frame.text = frame:CreateFontString(
      "$parent_FontString",
      "OVERLAY",
      "CombatLogFont"
    )
    frame.text:SetPoint("LEFT")
    frame.text:SetTextColor(1, 1, 1, 1)
    
    frame.raidMarker = CreateFrame("Frame", nil, SeedRaid.frame)
    frame.raidMarker.texture = frame.raidMarker:CreateTexture()
  else
    frame:ClearAllPoints()
  end
  frame:Show()
  return frame
end

--
-- Compares a given name to the player's name.
-- @notes
-- Never use to compare with SeedRaid.member entry. Only compare with
-- results returned from an API call.
function SeedRaid:IsPlayerName(name)
  if not name then
    return false
  end
  local playerName = GetUnitName("player", true)
  if name == playerName then
    return true
  end
  return false
end

--
-- Records what loot has been receieved.
--
function SeedRaid:LootReceieved(...)
  local message = select(1, ...)
  local name = select(5, ...)

  if (not message or type(message) ~= "string") or
     (not name    or type(name)    ~= "string") then
    return false
  end

  local loot = string.match(message, "%[(.+)%]")
  local count = string.match(message, "x(%d+)%.") or 1

  if not loot then
    return false
  end

  count = tonumber(count)

  if SeedRaid:IsPlayerName(name) then
    name = SeedRaid.playerName
  end

  if not name or SeedRaid.members[name] == nil then
    print(
      "SeedRaid:LootReceieved",
      "Error!", name, loot, count, "NOT RECORDED!"
    )
    return false
  end

  local member = SeedRaid.members[name]
  
  if member.loots[loot] == nil then
    member.loots[loot] = count
  else
    member.loots[loot] = member.loots[loot] + count
  end

  if SeedRaid.lootFrames[loot] == nil then
    SeedRaid.lootFrames[loot] = SeedRaid:GetCheckButton("loot")
    SeedRaid.lootFrames[loot]:Hide()
    SeedRaid.lootFrames[loot]:Disable()
  end

  if SeedRaid.frameSelected == member.frameRoster then
    return true
  end

  return false
end

--
-- Reacts to registered game events.
-- @param event The event that has been registered to be receieved.
-- @param ... Event specific arguements.
-- 
function SeedRaid:OnEvent(event, ...)
  local updateDisplay = false
  if event == "GROUP_ROSTER_UPDATE"   or
     event == "PARTY_LEADER_CHANGED"  or
     event == "PLAYER_ENTERING_WORLD" or 
     event == "UNIT_NAME_UPDATE"      then
     SeedRaid:UpdateMembers()
     SeedRaid:UpdateIcons()
     updateDisplay = true

  elseif event == "RAID_TARGET_UPDATE" then
    SeedRaid:UpdateIcons()
    updateDisplay = true

  elseif event == "UNIT_SPELLCAST_FAILED" then
    if SeedRaid.debug then
      updateDisplay = SeedRaid:SeedPlanted(...)
    end

  elseif event == "UNIT_SPELLCAST_SUCCEEDED" then
    updateDisplay = SeedRaid:SeedPlanted(...)

  elseif event == "CHAT_MSG_LOOT" then
    updateDisplay = SeedRaid:LootReceieved(...)

  end

  if updateDisplay then
    local success, values = xpcall(
      function()
        SeedRaid:UpdateRaidDisplay()
      end,
      function(msg)
        return {[1]=msg, [2]=debugstack()}
      end
    )
    if not success then
      print(values[1], values[2])
    end
  end
end

--
-- Scrolls a list when mousewheel is scrolled.
--
function SeedRaid.OnMouseWheel(self, delta)
  local scrollbar = SeedRaid.scrollbar
      if self == SeedRaid.seedClip or self == SeedRaid.scrollbarPlanted then
      scrollbar = SeedRaid.scrollbarPlanted
  elseif self == SeedRaid.lootClip or self == SeedRaid.scrollbarLoot    then
      scrollbar = SeedRaid.scrollbarLoot
  end

  local value = scrollbar:GetValue()
  local newvalue = value + delta * scrollbar.scrollStep * -1
  scrollbar:SetValue(newvalue)
  SeedRaid:UpdateRaidDisplay()
end

--
-- Moves CheckButton widget back to the frame pool.
--
function SeedRaid:RemoveCheckButton(frame, type)
  if type ~= "raid" and type ~= "seed" and type ~= "loot" then
    print(
      "SeedRaid:RemoveCheckButton",
      "Error, " .. tostring(type) .. " is not a valid framepool."
    )
    return
  end

  local pool = SeedRaid.framepool[type]

  if frame then
    frame:Hide()
    frame:SetChecked(false)
    frame.text:SetTextColor(1, 1, 1, 1)
    table.insert(pool, frame)
  else
    print(
      "SeedRaid:RemoveCheckButton",
      "Error, frame doesn't exist to remove."
    )
  end
end

--
-- Removes a raid member from the member table.
-- @param name Raid member name to remove.
--
function SeedRaid:RemoveMember(name)
  if not name or type(name) ~= "string" then
    return
  end

  local raidMember = SeedRaid.members[name]
  if raidMember ~= nil then
    SeedRaid:RemoveCheckButton(raidMember.frameRoster, "raid")
    SeedRaid.members[name] = nil
    SeedRaid.memberCount = SeedRaid.memberCount - 1
  else
    print(
      "SeedRaid:RemoveMember",
      "Error, " .. name .. "doesn't exist in table!"
    )
  end
end

-- 
-- Detects when seeds are planted.
-- @param ... The data received with the UNIT_SPELLCAST_SUCCEEDED event.
-- @return boolean Whether or not a seed was planted & recorded.
-- 
function SeedRaid:SeedPlanted(...)
  local unitId = select(1, ...)
  local lineId = select(4, ...)
  local spellId = select(5, ...)

  if (not unitId  or type(unitId)  ~= "string") or
     (not spellId or type(spellId) ~= "number") then
    return false
  end

  if not string.find(unitId, "party")  and
     not string.find(unitId, "player") and
     not string.find(unitId, "raid")   then
    return false
  end

  local name = nil
  
  if unitId == "player" and not IsInRaid() then
    name = SeedRaid.playerName
  else
    if IsInRaid() then
      name = GetRaidRosterInfo(string.match(unitId, "raid(%d+)"))
    elseif IsInGroup() then
      name = GetRaidRosterInfo(string.match(unitId, "party(%d+)"))
    end
    if SeedRaid:IsPlayerName(name) then
      name = SeedRaid.playerName
    end
  end
  
  local found = false
  for seedId, data in pairs(SeedRaid.seeds) do
    if spellId == seedId then
      found = true
      break
    end
  end
  if found == false then
    -- Spell is not a seed spell
    return false
  end

  if not name or SeedRaid.members[name] == nil then
    print(
      "SeedRaid:SeedPlanted",
      "Error, no such raid member.", name, GetSpellInfo(spellId), "NOT RECORDED!"
    )
    return false
  end

  local member = SeedRaid.members[name]

  member.seeds[spellId] = member.seeds[spellId] + 1
  if spellId ~= 193801 then
    -- Excepting Felwort, count this seed planted towards total
    member.seedsSum = member.seedsSum + 1
  end

  if SeedRaid.player.rank ~= SeedRaid.RAID_LEADER          or
      (name == SeedRaid.playerName and not SeedRaid.debug) or
      not SeedRaid.plantChannel                            then
    return true
  end

  if member.plantedAll then
    return true
  elseif member.seedsSum == SeedRaid.seedCount then
    member.plantedAll = true
    SendChatMessage(
      name .. ": " .. member.seedsSum .. ", DONE PLANTING.",
      SeedRaid.plantChannel,
      nil,
      name
    )
  elseif (member.seedsSum % SeedRaid.roundSize) == 0 and member.seedsSum > 0 then
    SendChatMessage(
      name .. ": " .. member.seedsSum .. ", done with round.",
      SeedRaid.plantChannel,
      nil,
      name
    )
  elseif (member.seedsSum % SeedRaid.alertInterval) == 0 and member.seedsSum > 0 then
    SendChatMessage(
      name .. ": " .. member.seedsSum .. " planted.",
      SeedRaid.plantChannel,
      nil,
      name
    )
  end

  return true
end

function SeedRaid:SelectPlayer()
  if not SeedRaid or not SeedRaid.members or not SeedRaid.playerName then
    return
  end
  SeedRaid.frameSelected = SeedRaid.members[SeedRaid.playerName].frameRoster
  SeedRaid.frameSelected:SetChecked(true)
end

--
-- Ensures only one raid member is selected at a time.
-- @note Should only be called from UI OnEvent. Otherwise, logging will break.
--
function SeedRaid:SetRaidRosterSelection(frame)
  if SeedRaid.frameSelected and SeedRaid.frameSelected == frame then
    -- Don't uncheck if a different entry wasn't selected
    SeedRaid.frameSelected:SetChecked(true)
    return
  end

  SeedRaid.frameSelected = frame

  if not SeedRaid.members then
    return
  end

  for member, data in pairs(SeedRaid.members) do
    if data.frameRoster and data.frameRoster ~= frame then
      data.frameRoster:SetChecked(false)
    end
  end

  if not SeedRaid.lootFrames then
    return
  end

  -- Hide all loot frames
  for loot, frame in pairs(SeedRaid.lootFrames) do
    frame:Hide()
  end

  SeedRaid:UpdateRaidDisplay()
end

--
-- Sets raid seed count when enter/escape pressed on seed count editbox.
--
function SeedRaid.SetSeedCount(self)
  self:ClearFocus()
  local num = self:GetNumber()
  if num > 0 then
    SeedRaid.seedCount = num
  end
  SeedRaid.roundSize = math.min(SeedRaid.roundSize, SeedRaid.seedCount)
  SeedRaid.alertInterval = math.min(SeedRaid.alertInterval, SeedRaid.roundSize)

  SR_SeedCount:SetText(SeedRaid.seedCount)
  SR_RoundSize:SetText(SeedRaid.roundSize)
  SR_AlertInterval:SetText(SeedRaid.alertInterval)
end

--
-- Sets round seed count when enter/escape pressed on round size editbox.
--
function SeedRaid.SetRoundSize(self)
  self:ClearFocus()
  local num = self:GetNumber()
  if num > 0 then
    SeedRaid.roundSize = num
  end
  SeedRaid.seedCount = math.max(SeedRaid.seedCount, SeedRaid.roundSize)
  SeedRaid.alertInterval = math.min(SeedRaid.alertInterval, SeedRaid.roundSize)

  SR_SeedCount:SetText(SeedRaid.seedCount)
  SR_RoundSize:SetText(SeedRaid.roundSize)
  SR_AlertInterval:SetText(SeedRaid.alertInterval)
end

--
-- Sets alert seed count when enter/escape pressed on alert interval editbox.
--
function SeedRaid.SetAlertInterval(self)
  self:ClearFocus()
  local num = self:GetNumber()
  if num > 0 then
    SeedRaid.alertInterval = num
  end
  SeedRaid.seedCount = math.max(SeedRaid.seedCount, SeedRaid.alertInterval)
  SeedRaid.roundSize = math.max(SeedRaid.roundSize, SeedRaid.alertInterval)

  SR_SeedCount:SetText(SeedRaid.seedCount)
  SR_RoundSize:SetText(SeedRaid.roundSize)
  SR_AlertInterval:SetText(SeedRaid.alertInterval)
end

-- 
-- Determines which raid members are currently marked for planting.
-- @notes
-- This only works in a raid, not in a party.
-- 0 = None
-- 1 = Yellow 4-point Star
-- 2 = Orange Circle
-- 3 = Purple Diamond
-- 4 = Green Triangle
-- 5 = White Crescent Moon
-- 6 = Blue Square
-- 7 = Red "X" Cross
-- 8 = White Skull
-- 
function SeedRaid:UpdateIcons()
  if not SeedRaid.members then
    return
  end
  for member, data in pairs(SeedRaid.members) do
    if data.inPartyOrRaid then
      data.marked = GetRaidTargetIndex(data.unitId) or 0
    else
      data.marked = 0
    end
  end
end

--
-- Gets the name of every raid member and initializes their data structure.
-- @notes
-- GetRaidRosterInfo works for both raids and parties. 
-- If the member name isn't yet known, GetRaidRosterInfo will return nil.
-- GetUnitName will return the valid text "Unknown". This could be filtered,
-- but it gets translated for every language available, which is cumbersome
-- to check for in code.
-- The important events to trigger this function are:
-- GROUP_ROSTER_UPDATE, PLAYER_ENTERING_WORLD, UNIT_NAME_UPDATE
--
-- Not using WoW constant MAX_PARTY_MEMBERS, as that is 4 not 5. It was made
-- before CompactRaidFrames was a thing, so you had the player frame in upper
-- left corner and then 4 other party members frames shown when in a group.
--
function SeedRaid:UpdateMembers()
  if not SeedRaid.members then
    return
  end

  for name, data in pairs(SeedRaid.members) do
    if name ~= SeedRaid.playerName then
      data.inPartyOrRaid = false
    end
    data.rank = SeedRaid.RAID_MEMBER
  end

  local limit = 5
  local unit = "party"
  if IsInRaid() then
    limit = 40
    unit = "raid"
  end

  local num = GetNumGroupMembers()
  for i=1, limit do
    if i <= num then
      local name, rank = GetRaidRosterInfo(i)
      rank = rank or SeedRaid.RAID_MEMBER
      if name then
        if SeedRaid:IsPlayerName(name) then
          name = SeedRaid.playerName
        end
        local member = SeedRaid:AddMember(name, unit..i, rank)
        if member then
          member.inPartyOrRaid = true
        end
      end
    end
  end
end

--
-- Match the UI to Lua data
--
function SeedRaid:UpdateRaidDisplay()
  -- Titlebar

  local memberCount = 0
  for name, data in pairs(SeedRaid.members) do
    if data.inPartyOrRaid then
      memberCount = memberCount + 1
    end
  end

  SR_Title:SetText("Seed Raid " .. memberCount .. "/10")
  SR_Title:SetPoint("LEFT", 5, 0)
  SR_SeedCountText:SetPoint("LEFT", SR_Title, "RIGHT", 20, 0)
  SR_SeedCount:SetPoint("LEFT", SR_SeedCountText, "RIGHT", 5, 0)
  SR_RoundSizeText:SetPoint("LEFT", SR_SeedCount, "RIGHT", 20, 0)
  SR_RoundSize:SetPoint("LEFT", SR_RoundSizeText, "RIGHT", 5, 0)
  SR_AlertIntervalText:SetPoint("LEFT", SR_RoundSize, "RIGHT", 20, 0)
  SR_AlertInterval:SetPoint("LEFT", SR_AlertIntervalText, "RIGHT", 5, 0)

  -- Reset SR frame width
  local left, bottom, width, height = SR_AlertInterval:GetRect()
  local right = left + width + 30
  local left, bottom, width, height = SR:GetRect()
  local width = right - left
  SR:SetWidth(width)
  
  -- Player List

  local colorFinished = {0.0, 0.5, 0.505, 1.0}
  local colorNormal = {1, 1, 1, 1}
  local colorNotInPartyOrRaid = {0.45, 0.45, 0.45, 1.0}

  -- Re-flow the layout based on current raid status (SeedRaid.comparePairs)
  local count = 1
  local previous = nil
  for name, data in SeedRaid:orderedPairs(SeedRaid.members) do
    data.frameRoster:Show()
    if count == 1 then
      data.frameRoster:SetPoint("TOPLEFT")
    else
      data.frameRoster:SetPoint("TOPLEFT", previous, "BOTTOMLEFT")
    end

    local text = string.format("%3.0f", data.seedsSum) .. " " .. name
    data.frameRoster.text:SetText(text)
    if not data.inPartyOrRaid then
      data.frameRoster.text:SetTextColor(unpack(colorNotInPartyOrRaid))
    elseif data.seedsSum >= SeedRaid.seedCount then
      data.frameRoster.text:SetTextColor(unpack(colorFinished))
    else
      data.frameRoster.text:SetTextColor(unpack(colorNormal))
    end

    if data.marked > 0 then
      local mark = tostring(data.marked)
      local file = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_" .. mark
      data.frameRoster.raidMarker.texture:SetTexture(file);
      data.frameRoster.raidMarker.texture:SetAllPoints()
      data.frameRoster.raidMarker:SetHeight(16)
      data.frameRoster.raidMarker:SetWidth(16)
      data.frameRoster.raidMarker:SetPoint("RIGHT", data.frameRoster, "LEFT")
      data.frameRoster.raidMarker:Show()
    else
      data.frameRoster.raidMarker:Hide()
    end

    count = count + 1
    previous = data.frameRoster
  end
  
  -- Adjust scrollbar
  if previous then
    local entryHeight = previous:GetHeight()
    local contentSize = entryHeight * (count - 1)
    local scrollSize = SeedRaid.raidClip:GetHeight()
    local scroll = contentSize - scrollSize
    if scroll <= 0 then
      SeedRaid.scrollbar:Hide()
    else
      SeedRaid.scrollbar:Show()
    end
    scroll = math.max(1, scroll)
    SeedRaid.scrollbar:SetMinMaxValues(1, scroll)
    SeedRaid.scrollbar:SetValueStep(entryHeight)
    SeedRaid.scrollbar.scrollStep = entryHeight

    local scrollBottom = SeedRaid.raidClip:GetBottom()
    local scrollTop = scrollBottom + scrollSize
    -- Hide players outside of scroll area
    for name, data in pairs(SeedRaid.members) do
      local left, bottom, width, height = data.frameRoster:GetBoundsRect()
      if (bottom > scrollTop)             or
        (bottom + height < scrollBottom) then
        data.frameRoster:Hide()
        data.frameRoster.raidMarker:Hide()
      end
    end
  else
    SeedRaid.scrollbar:Hide()
  end

  -- Seed List

  local selected = SeedRaid.frameSelected.text:GetText()
  local name = string.match(selected, "%d+ (.*)")
  local member = SeedRaid.members[name]
  if member == nil then
    return
  end

  for seedId, count in pairs(member.seeds) do
    local text = "" .. count .. " " .. SeedRaid.seeds[seedId].name
    local frame = SeedRaid.seeds[seedId].frame
    frame.text:SetText(text)
    frame:Show()
  end

  SeedRaid.scrollbarPlanted:Hide()

  -- Loot List

  local count = 1
  local previous = nil
  for loot, lootCount in SeedRaid:orderedPairs(member.loots) do
    local frame = SeedRaid.lootFrames[loot]
    frame:Show()
    if count > 1 then
      frame:SetPoint("TOPLEFT", previous, "BOTTOMLEFT")
    else
      frame:SetPoint("TOPLEFT")
    end
    
    local text = "" .. lootCount .. " " .. loot
    frame.text:SetText(text)

    previous = frame
    count = count + 1
  end
  
  -- Adjust scrollbar
  if previous then
    local entryHeight = previous:GetHeight()
    local contentSize = entryHeight * (count - 1)
    local scrollSize = SeedRaid.lootClip:GetHeight()
    local scroll = contentSize - scrollSize
    if scroll <= 0 then
      SeedRaid.scrollbarLoot:Hide()
    else
      SeedRaid.scrollbarLoot:Show()
    end
    scroll = math.max(1, scroll)
    SeedRaid.scrollbarLoot:SetMinMaxValues(1, scroll)
    SeedRaid.scrollbarLoot:SetValueStep(entryHeight)
    SeedRaid.scrollbarLoot.scrollStep = entryHeight

    local scrollBottom = SeedRaid.lootClip:GetBottom()
    local scrollTop = scrollBottom + scrollSize
    -- Hide loot outside of scroll area
    for loot, lootCount in pairs(member.loots) do
      local frame = SeedRaid.lootFrames[loot]
      local left, bottom, width, height = frame:GetBoundsRect()
      if (bottom > scrollTop)             or
         (bottom + height < scrollBottom) then
        frame:Hide()
      end
    end
  else
    SeedRaid.scrollbarLoot:Hide()
  end
end

-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------

--
-- Returns a table iterator, which is sorted by the table key.
-- @param t The table to iterate.
-- 
function SeedRaid:orderedPairs(t)
  local ordered = {}
  for key, value in pairs(t) do
    table.insert(ordered, {[1]=key,[2]=value})
  end
  table.sort(ordered, SeedRaid.comparePairs)
  local i = 0
  local n = table.getn(ordered)

  return function()
    i = i + 1
    if i <= n then
      local pair = ordered[i]
      return pair[1], pair[2]
    end
    return nil
  end
end

function SeedRaid.comparePairs(a, b)
  if a[1] == SeedRaid.playerName then
    return true
  elseif b[1] == SeedRaid.playerName then
    return false
  elseif type(a[2]) == "table" and 
         type(b[2]) == "table" then
    if a[2].marked > 0 and b[2].marked == 0 then
      return true
    elseif b[2].marked > 0 and a[2].marked == 0 then
      return false
    elseif a[2].inPartyOrRaid and not b[2].inPartyOrRaid then
      return true
    elseif b[2].inPartyOrRaid and not a[2].inPartyOrRaid then
      return false
    end
  end

  return a[1] < b[1]
end

-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------

--
-- Resets all data, just like a /reload would.
--
function SeedRaid:Reset()
  SeedRaid.scrollbar:SetValue(1) 
  SeedRaid.scrollbar:SetMinMaxValues(1, 1)
  SeedRaid.scrollbar:SetWidth(16)

  SeedRaid.scrollbarPlanted:SetValue(1) 
  SeedRaid.scrollbarPlanted:SetMinMaxValues(1, 1)
  SeedRaid.scrollbarPlanted:SetWidth(16)

  SeedRaid.scrollbarLoot:SetValue(1) 
  SeedRaid.scrollbarLoot:SetMinMaxValues(1, 1)
  SeedRaid.scrollbarLoot:SetWidth(16)

  SeedRaid:DisableEvents()

  for name, data in pairs(SeedRaid.members) do
    SeedRaid:RemoveMember(name)
  end

  -- Hide all loot frames
  for loot, frame in pairs(SeedRaid.lootFrames) do
    frame:Hide()
  end

  SeedRaid.playerName = nil
  SeedRaid.player = nil

  -- Add player
  SeedRaid:AddPlayer()

  -- Set player as default selection
  SeedRaid:SelectPlayer()

  -- Add current raid members
  SeedRaid:UpdateMembers()
  SeedRaid:UpdateIcons()
  SeedRaid:UpdateRaidDisplay()

  SeedRaid:EnableEvents()
end

SeedRaid.RAID_MEMBER = 0
SeedRaid.RAID_ASSISTANT = 1
SeedRaid.RAID_LEADER = 2

-- Register ESC to close SeedRaid frame
table.insert(UISpecialFrames, "SR") 

-- Register mouse events with SeedRaid frame
SeedRaid.frame:RegisterForDrag("LeftButton")

-- Register mousewheel with clip frames
SeedRaid.raidClip:EnableMouseWheel(true)
SeedRaid.scrollbar:EnableMouseWheel(true)
SeedRaid.seedClip:EnableMouseWheel(true)
SeedRaid.scrollbarPlanted:EnableMouseWheel(true)
SeedRaid.lootClip:EnableMouseWheel(true)
SeedRaid.scrollbarLoot:EnableMouseWheel(true)

-- Register game events with SeedRaid frame
SeedRaid.frame:RegisterEvent("ADDON_LOADED")
SeedRaid.frame:RegisterEvent("GROUP_ROSTER_UPDATE")
SeedRaid.frame:RegisterEvent("RAID_TARGET_UPDATE")
SeedRaid.frame:RegisterEvent("CHAT_MSG_LOOT")
SeedRaid.frame:RegisterEvent("PARTY_LEADER_CHANGED")
SeedRaid.frame:RegisterEvent("PLAYER_ENTERING_WORLD")
SeedRaid.frame:RegisterEvent("UNIT_NAME_UPDATE")
SeedRaid.frame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
SeedRaid.frame:RegisterEvent("UNIT_SPELLCAST_FAILED")

-- One time setup
SeedRaid.framepool = {["raid"]={}, ["seed"]={}, ["loot"]={}}
SeedRaid.framepoolParents = {
  ["raid"]=SeedRaid.content,
  ["seed"]=SeedRaid.contentPlanted,
  ["loot"]=SeedRaid.contentLoot
}
SeedRaid.members = {}
SeedRaid.memberCount = 0
SeedRaid.reportChannel = print
SeedRaid.plantChannel = "WHISPER"
SeedRaid.seeds = {
  [193795] = {["name"]="Aethril",        ["frame"]=SeedRaid:GetCheckButton("seed")},
  [193797] = {["name"]="Dreamleaf",      ["frame"]=SeedRaid:GetCheckButton("seed")},
  [193801] = {["name"]="Felwort",        ["frame"]=SeedRaid:GetCheckButton("seed")},
  [193799] = {["name"]="Fjarnskaggl",    ["frame"]=SeedRaid:GetCheckButton("seed")},
  [193798] = {["name"]="Foxflower",      ["frame"]=SeedRaid:GetCheckButton("seed")},
  [193800] = {["name"]="Starlight Rose", ["frame"]=SeedRaid:GetCheckButton("seed")}
}
SeedRaid.seeds[193795].frame:SetPoint("TOPLEFT")
SeedRaid.seeds[193797].frame:SetPoint("TOPLEFT", SeedRaid.seeds[193795].frame, "BOTTOMLEFT")
SeedRaid.seeds[193801].frame:SetPoint("TOPLEFT", SeedRaid.seeds[193797].frame, "BOTTOMLEFT")
SeedRaid.seeds[193799].frame:SetPoint("TOPLEFT", SeedRaid.seeds[193801].frame, "BOTTOMLEFT")
SeedRaid.seeds[193798].frame:SetPoint("TOPLEFT", SeedRaid.seeds[193799].frame, "BOTTOMLEFT")
SeedRaid.seeds[193800].frame:SetPoint("TOPLEFT", SeedRaid.seeds[193798].frame, "BOTTOMLEFT")
for i, seed in pairs(SeedRaid.seeds) do
  seed.frame:Disable()
end

SeedRaid.lootFrames = {}

SeedRaid.playerName = nil
SeedRaid.player = nil
SeedRaid.frameSelected = nil

-- Set CheckButton sizes
local height = 15
-- SR_EntryTemplate:SetHeight(height)
SeedRaid.raidClip:SetHeight(height*11)
SeedRaid.content:SetHeight(height*40)

-- Default titlebar values
SeedRaid.seedCount = 100
SeedRaid.roundSize = 50
SeedRaid.alertInterval = 10

SR_SeedCount:SetText(SeedRaid.seedCount)
SR_RoundSize:SetText(SeedRaid.roundSize)
SR_AlertInterval:SetText(SeedRaid.alertInterval)

-- Create minimap icon
local iconObject = LibStub("LibDataBroker-1.1"):NewDataObject("Seed Raid", {
	type = "launcher",
	text = "Seed Raid",
	icon = "656679",
	OnClick =
  function(...)
    -- select(2, ...) LeftButton RightButton MiddleButton Button4 Button5
    local button = select(2, ...)
    if button == "LeftButton" then
      if SeedRaid.frame:IsVisible() then
        SeedRaid.frame:Hide()
      else
        SeedRaid.frame:Show()
      end
    elseif button == "MiddleButton" then
      if SeedRaid.plantChannel == "WHISPER" then
        SeedRaid.plantChannel = nil
        print("SeedRaid whispering disabled.")
      else
        SeedRaid.plantChannel = "WHISPER"
        print("SeedRaid whispering enabled.")
      end
    elseif button == "RightButton" then
      SeedRaid:Reset()
      print("SeedRaid results cleared.")
    end
  end
})
local icon = LibStub("LibDBIcon-1.0")

SeedRaid.icon = {hide = false}
icon:Register("Seed Raid", iconObject, SeedRaid.icon)

-- First time setup
SeedRaid:Reset()